
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2013 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layApplication.h"
#include "layConfig.h"
#include "layMainWindow.h"
#include "layVersion.h"
#include "layExceptions.h"
#include "layMacro.h"
#include "gtf.h"
#include "gsiDecl.h"
#include "rdb.h"
#include "dbStatic.h"
#include "dbLibrary.h"
#include "dbLibraryManager.h"
#include "tlException.h"
#include "tlAssert.h"
#include "tlLog.h"
#include "tlStaticObjects.h"
#include "tlString.h"
#include "tlExpression.h"
#include "tlInternational.h"

#include <QtGui/QIcon>
#include <QtCore/QDir>
#include <QtCore/QFileInfo>
#include <QtCore/QFile>
#include <QtGui/QAction>

#ifdef _WIN32
#  include <windows.h>
//  get rid of these - we have std::min/max ..
#  ifdef min
#    undef min
#  endif
#  ifdef max
#    undef max
#  endif
#else
#include <dlfcn.h>
#endif

#include <iostream>
#include <memory>
#include <algorithm>

namespace lay
{

// --------------------------------------------------------------------------------

#if HAVE_RUBY
const bool s_ruby_available = true;
#else
const bool s_ruby_available = false;
#endif

static Application *ms_instance = 0;

static void load_plugin (const std::string &pp)
{
#ifdef _WIN32
  //  there is no "dlopen" on mingw, so we need to emulate it.
  HINSTANCE handle = LoadLibraryW ((const wchar_t *) tl::to_qstring (pp).constData ());
  if (! handle) {
    throw tl::Exception (tl::translate ("Unable to load plugin: %s with error message: %s "), pp, GetLastError ());
  }
#else
  void *handle;
  handle = dlopen (tl::string_to_system (pp).c_str (), RTLD_LAZY);
  if (! handle) {
    throw tl::Exception (tl::translate ("Unable to load plugin: %s"), pp);
  }
#endif

  tl::log << "Loaded plugin '" << pp << "'";
}

static void
split_path (const std::string &path, std::vector <std::string> &pc)
{
  QString sep;
#ifdef _WIN32
  sep = QString::fromAscii (";");
#else
  sep = QString::fromAscii (":");
#endif

  QStringList pp = tl::to_qstring (path).split (sep, QString::SkipEmptyParts);
  for (QStringList::ConstIterator p = pp.begin (); p != pp.end (); ++p) {
    pc.push_back (tl::to_string (*p));
  }
}

static std::string 
get_appdata_path ()
{
  QDir appdata_dir = QDir::homePath ();
  QString appdata_folder;
#ifdef _WIN32
  appdata_folder = QString::fromAscii ("KLayout");
#else
  appdata_folder = QString::fromAscii (".klayout");
#endif

  //  create the basic folder hierarchy
  if (! appdata_dir.exists (appdata_folder)) {
    appdata_dir.mkdir (appdata_folder);
  }

  QString appdata_klayout_path = appdata_dir.absoluteFilePath (appdata_folder);
  QDir appdata_klayout_dir (appdata_klayout_path);

  QString macros_folder = QString::fromAscii ("macros");
  if (! appdata_klayout_dir.exists (macros_folder)) {
    appdata_klayout_dir.mkdir (macros_folder);
  }

  QString libs_folder = QString::fromAscii ("libraries");
  if (! appdata_klayout_dir.exists (libs_folder)) {
    appdata_klayout_dir.mkdir (libs_folder);
  }

  return tl::to_string (appdata_klayout_path);
}

static void 
get_other_system_paths (std::vector <std::string> &p)
{
#ifdef _WIN32
  //  Use "Application Data" if it exists on Windows
  const wchar_t *appdata_env = 0;
  if ((appdata_env = _wgetenv (L"APPDATA")) != 0) {
    QDir appdata_dir = QString ((const QChar *) appdata_env);
    QString appdata_folder = QString::fromAscii ("KLayout");
    if (appdata_dir.exists () && appdata_dir.exists (appdata_folder)) {
      p.push_back (tl::to_string (appdata_dir.absoluteFilePath (appdata_folder)));
    }
  }
#endif
}

Application::Application (int &argc, char *argv [], bool non_ui_mode)
  : QApplication (argc, argv, !non_ui_mode),
    m_lyp_map_all_cvs (true), 
    m_lyp_add_default (false),
    m_write_config_file (true),
    m_gtf_replay_rate (0),
    m_gtf_replay_stop (-1),
    m_same_view (false),
    m_sync_mode (false),
    m_no_gui (false),
    m_editable (false),
    m_editable_set (false),
    m_enable_undo (true),
    mp_qapp (0),
    mp_qapp_gui (0),
    mp_ruby_interpreter (0),
    mp_mw (0),
    mp_plugin_root (0),
    mp_recorder (0)
{
  mp_qapp = this;
  mp_qapp_gui = (non_ui_mode ? 0 : this);
  
  //  initialize the system codecs (Hint: this must be done after the QApplication is initialized because
  //  it will call setlocale)
  tl::initialize_codecs ();

  //  transscribe the arguments to UTF8
  std::vector<std::string> args;
  args.reserve (argc);
  for (int i = 0; i < argc; ++i) {
    args.push_back (argv [i]);
  }

  tl_assert (ms_instance == 0);
  ms_instance = this;

  std::string gtf_record;
  bool gtf_save_incremental = false;

  //  get and create the klayout appdata folder if required
  m_appdata_path = get_appdata_path ();

  //  get the installation path
  m_inst_path = tl::to_string (QCoreApplication::applicationDirPath ());

  //  generate the klayout path: the first component is always the appdata path
  m_klayout_path.push_back (m_appdata_path);
#ifdef _WIN32
  wchar_t *klayout_path = _wgetenv (L"KLAYOUT_PATH");
  if (klayout_path) {
    split_path (tl::to_string (QString ((const QChar *) klayout_path)), m_klayout_path);
  } else {
    get_other_system_paths (m_klayout_path);
    m_klayout_path.push_back (m_inst_path);
  }
#else
  char *klayout_path = getenv ("KLAYOUT_PATH");
  if (klayout_path) {
    split_path (tl::system_to_string (klayout_path), m_klayout_path);
  } else {
    get_other_system_paths (m_klayout_path);
    m_klayout_path.push_back (m_inst_path);
  }
#endif

  if (mp_qapp_gui) {

    //  create the configuration files paths and collect the initialization config files
    //  (the ones used for reset) into m_initial_config_files.
    {
      //  Fallback to ~/.layviewrc for backward compatibility
      QDir qd (QDir::home ());
      QString filename = QString::fromAscii (".layviewrc");
      if (qd.exists (filename) && QFileInfo (qd.absoluteFilePath (filename)).isReadable ()) {
        m_config_files.push_back (tl::to_string (qd.absoluteFilePath (filename)));
        m_config_file_to_delete = m_config_files.back ();
      }
    }

    m_config_file_to_write = tl::to_string (QDir (tl::to_qstring (m_appdata_path)).absoluteFilePath (QString::fromAscii ("klayoutrc")));

    //  Hint: the order is reverse in the sense that the first one wins ...
    for (std::vector <std::string>::const_iterator p = m_klayout_path.end (); p != m_klayout_path.begin (); ) {
      --p;
      QDir qd (tl::to_qstring (*p));
      QString filename = QString::fromAscii ("klayoutrc");
      if (qd.exists (filename) && QFileInfo (qd.absoluteFilePath (filename)).isReadable ()) {
        m_config_files.push_back (tl::to_string (qd.absoluteFilePath (filename)));
        if (m_config_files.back () != m_config_file_to_write) {
          m_initial_config_files.push_back (m_config_files.back ());
        }
      }
    }

  }

  //  try to locate a global rbainit file and rbm modules
  std::vector<std::string> global_rba_modules;
  if (s_ruby_available) {

    std::set<std::string> modules;

    //  try to locate a global plugins
    for (std::vector <std::string>::const_iterator p = m_klayout_path.begin (); p != m_klayout_path.end (); ++p) {

      QFileInfo rbainit_file (tl::to_qstring (*p), QString::fromAscii ("rbainit"));
      if (rbainit_file.exists () && rbainit_file.isReadable ()) {
        std::string m = tl::to_string (rbainit_file.absoluteFilePath ());
        if (modules.find (m) == modules.end ()) {
          global_rba_modules.push_back (m);
          modules.insert (m);
        }
      }

      QDir inst_path_dir (tl::to_qstring (*p));

      QStringList name_filters;
      name_filters << QString::fromAscii ("*.rbm");

      QStringList inst_modules = inst_path_dir.entryList (name_filters);
      inst_modules.sort ();

      for (QStringList::const_iterator im = inst_modules.begin (); im != inst_modules.end (); ++im) {
        QFileInfo rbm_file (tl::to_qstring (*p), *im);
        if (rbm_file.exists () && rbm_file.isReadable ()) {
          std::string m = tl::to_string (rbm_file.absoluteFilePath ());
          if (modules.find (m) == modules.end ()) {
            global_rba_modules.push_back (m);
            modules.insert (m);
          }
        }
      }

    }

  }

  //  try to locate the global plugins
  for (std::vector <std::string>::const_iterator p = m_klayout_path.begin (); p != m_klayout_path.end (); ++p) {

    std::set<std::string> modules;

    QDir inst_path_dir (tl::to_qstring (*p));

    QStringList name_filters;
    name_filters << QString::fromAscii ("*.klp");

    QStringList inst_modules = inst_path_dir.entryList (name_filters);
    inst_modules.sort ();

    for (QStringList::const_iterator im = inst_modules.begin (); im != inst_modules.end (); ++im) {
      QFileInfo klp_file (tl::to_qstring (*p), *im);
      if (klp_file.exists () && klp_file.isReadable ()) {
        std::string m = tl::to_string (klp_file.absoluteFilePath ());
        if (modules.find (m) == modules.end ()) {
          load_plugin (m);
          modules.insert (m);
        }
      }
    }

  }

  std::vector <std::string> custom_macro_paths;
  bool no_macros = false;

  // currently: technology is always set to make "default" technology the default
  bool tech_set = true; 
  std::string tech;

  for (int i = 1; i < argc; ++i) {

    const std::string &a = args [i];

    if (a == "-d" && (i + 1) < argc) {

      int v = 0;
      tl::from_string (args [++i], v);
      v = std::max (0, v);
      tl::verbosity (v);

    } else if (a == "-l" && (i + 1) < argc) {

      m_layer_props_file = args [++i];

    } else if (a == "-lx") {

      m_lyp_add_default = true;

    } else if (a == "-lf") {

      m_lyp_map_all_cvs = false;

    } else if (a == "-u" && (i + 1) < argc) {

      m_session_file = args [++i];

    } else if (a == "-wd" && (i + 1) < argc) {

      std::string v;
      const char *p = args [++i].c_str ();
      const char *n0 = p;
      while (*p && *p != '=') {
        ++p;
      }
      std::string n (n0, 0, p - n0);
      if (*p == '=') {
        tl::Eval::set_global_var (n, tl::Variant (v));
      } else {
        tl::Eval::set_global_var (n, tl::Variant (true));
      }

    } else if (s_ruby_available && a == "-rd" && (i + 1) < argc) {

      std::string v;
      const char *p = args [++i].c_str ();
      const char *n0 = p;
      while (*p && *p != '=') {
        ++p;
      }
      std::string n (n0, 0, p - n0);
      m_variables.push_back (std::make_pair (n, v));
      if (*p == '=') {
        m_variables.back ().second = ++p;
      }

    } else if (s_ruby_available && a == "-rm" && (i + 1) < argc) {

      m_ruby_modules.push_back (args [++i]);

    } else if (s_ruby_available && a == "-r" && (i + 1) < argc) {

      m_ruby_script = args [++i];

    } else if (s_ruby_available && a == "-rx") {

      no_macros = true;

    } else if (a == "-gr" && (i + 1) < argc) {

      gtf_record = args [++i];

    } else if (a == "-gi") {

      gtf_save_incremental = true;

    } else if (a == "-gp" && (i + 1) < argc) {

      m_gtf_replay = args [++i];
      //  test mode replay forces sync mode to true
      m_sync_mode = true;

    } else if (a == "-gx" && (i + 1) < argc) {

      int r = 0;
      tl::from_string (args [++i], r);
      m_gtf_replay_rate = std::max (0, r);

    } else if (a == "-gb" && (i + 1) < argc) {

      int s = 0;
      tl::from_string (args [++i], s);
      m_gtf_replay_stop = std::max (0, s);

    } else if (a == "-c" && (i + 1) < argc) {

      m_config_files.clear ();
      m_config_files.push_back (args [++i]);
      m_initial_config_files = m_config_files;
      m_config_file_to_write = m_config_files.back ();

    } else if (a == "-nc") {

      m_config_files.clear ();
      m_initial_config_files = m_config_files;
      m_config_file_to_write.clear ();

    } else if (a == "-n" && (i + 1) < argc) {

      tech = args [++i];
      tech_set = true;

    } else if (a == "-p" && (i + 1) < argc) {

      load_plugin (args [++i]);

    } else if (a == "-s") {

      m_same_view = true;
      
    } else if (a == "-e") {

      m_editable = true;
      m_editable_set = true;
      
    } else if (a == "-ne") {

      m_editable = false;
      m_editable_set = true;
      
    } else if (a == "-i") {

      m_enable_undo = false;
      
    } else if (a == "-ni") {

      m_enable_undo = true;
      
    } else if (a == "-j" && (i + 1) < argc) {

      custom_macro_paths.push_back (args [++i]);
      
    } else if (a == "-nt") {

      m_write_config_file = true;
      
    } else if (a == "-t") {

      m_write_config_file = false;
      
    } else if (a == "-z") {

      m_no_gui = true;
      
    } else if (a == "-zz") {

      m_no_gui = true;
      //  otherwise ignore (was set before)
      
    } else if (a == "-x") {

      m_sync_mode = true;
      
    } else if (a == "-v") {

      tl::info << lay::Version::name () << " " << lay::Version::version ();
      exit (0);
      
    } else if (a == "-h") {

      tl::info << usage () << tl::noendl;
      exit (0);

    } else if (a == "-m" && (i + 1) < argc) {

      m_files.push_back (std::make_pair (rdb_file, std::make_pair (std::string (args [++i]), std::string ())));

    } else if (a[0] == '-') {

      throw tl::Exception (tl::translate ("Unknown option: ") + a);

    } else {
      m_files.push_back (std::make_pair (tech_set ? layout_file_with_tech : layout_file, std::make_pair (a, tech)));
    }

  }

  //  create the ruby interpreter instance now.
  //  Hint: we do this after load_plugin, because that way the plugins can register GSI classes and methods.
  mp_ruby_interpreter = new rba::RubyInterpreter ();

  if (! no_macros) {
    //  Add the global ruby modules as the first ones.
    m_ruby_modules.insert (m_ruby_modules.begin (), global_rba_modules.begin (), global_rba_modules.end ());
  }

  //  Scan build-in macros
  //  These macros are always taken, even if there are no macros requested (they are required to 
  //  fully form the API).
  lay::MacroCollection::root ().add_folder (tl::translate ("Build-In"), ":/build-in-macros", true);

  if (! no_macros) {

    //  Scan for macros
    for (std::vector <std::string>::const_iterator p = m_klayout_path.begin (); p != m_klayout_path.end (); ++p) {
      std::string mp = tl::to_string (QDir (tl::to_qstring (*p)).filePath (tl::to_qstring ("macros")));
      if (p == m_klayout_path.begin ()) {
        lay::MacroCollection::root ().add_folder (tl::translate ("Local"), mp, false);
      } else if (m_klayout_path.size () == 2) {
        lay::MacroCollection::root ().add_folder (tl::translate ("Global"), mp, true);
      } else {
        lay::MacroCollection::root ().add_folder (tl::translate ("Global") + " - " + *p, mp, true);
      }
      ruby_interpreter ().add_path (mp);
    }

  }

  //  Read some configuration values that we need early
  bool editable_from_config = false;

  {
    lay::PluginRoot cfg;

    for (std::vector <std::string>::const_iterator c = m_config_files.begin (); c != m_config_files.end (); ++c) {
      try {
        cfg.read_config (*c);
      } catch (...) { }
    }

    try {
      cfg.config_get (cfg_edit_mode, editable_from_config);
    } catch (...) { }

    try {
      std::string mp;
      cfg.config_get (cfg_custom_macro_paths, mp);
      tl::Extractor ex (mp.c_str ());
      while (! ex.at_end ()) {
        std::string p;
        ex.read_word_or_quoted (p);
        custom_macro_paths.push_back (p);
        ex.test (";");
      }
    } catch (...) { }

  }

  //  Install the folders
  if (! no_macros) {
    for (std::vector <std::string>::const_iterator p = custom_macro_paths.begin (); p != custom_macro_paths.end (); ++p) {
      lay::MacroCollection::root ().add_folder (tl::translate ("Project") + " - " + *p, *p, false);
    }
  }

  //  If the editable flag was not set, use it from the 
  //  configuration. Since it is too early now, we cannot use the
  //  configuration once it is read
  if (! m_editable_set) {
    m_editable = editable_from_config;
  }

  db::set_default_editable_mode (m_editable);
  db::enable_transactions (m_enable_undo);

  if (mp_qapp_gui) {
    mp_qapp_gui->setWindowIcon (QIcon (QString::fromAscii (":/logo.png")));
#if QT_VERSION >= 0x040500
    mp_qapp_gui->setAttribute (Qt::AA_DontShowIconsInMenus, false);
#endif
  }

  if (mp_qapp_gui && ! gtf_record.empty ()) {
    //  since the recorder tracks QAction connections etc., it must be instantiated before every other 
    //  object performing a gtf::action_connect for example
    mp_recorder = new gtf::Recorder (mp_qapp_gui, gtf_record);
    mp_recorder->save_incremental (gtf_save_incremental);
  }

  tl::Eval::set_global_var ("appdata_path", tl::Variant (m_appdata_path));
  tl::Eval::set_global_var ("inst_path", tl::Variant (m_inst_path));

  tl::Variant kp (m_klayout_path.begin (), m_klayout_path.end ());
  tl::Eval::set_global_var ("klayout_path", kp);

  //  run all early autorun macros
  lay::MacroCollection::root ().autorun_early ();

  if (mp_qapp_gui) {
    mp_mw = new lay::MainWindow (mp_qapp_gui, "main_window");
    QObject::connect (mp_mw, SIGNAL (closed ()), mp_qapp_gui, SLOT (quit ()));
    mp_plugin_root = mp_mw;
  } else {
    mp_plugin_root = new lay::PluginRoot ();
  }

  //  initialize the plugins (this should be the last action in the constructor since the
  //  main window should be functional now.
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    lay::PluginDeclaration *pd = const_cast<lay::PluginDeclaration *> (&*cls);
    pd->initialize (mp_mw);
  }

  //  establish the configuration
  mp_plugin_root->config_setup ();

  //  Some info output 
  if (tl::verbosity () >= 20) {

    tl::info << "KLayout path:";
    for (std::vector <std::string>::const_iterator c = m_klayout_path.begin (); c != m_klayout_path.end (); ++c) {
      tl::info << "  " << *c;
    }
    tl::info << "Config file to write: " << m_config_file_to_write;
    tl::info << "Config files to read:";
    for (std::vector <std::string>::const_iterator c = m_config_files.begin (); c != m_config_files.end (); ++c) {
      tl::info << "  " << *c;
    }

  }
}

Application::~Application () 
{
  if (! ms_instance) {
    return;
  }

  shutdown ();
}

bool 
Application::notify (QObject *receiver, QEvent *e)
{
  bool ret = true;
  BEGIN_PROTECTED
    ret = QApplication::notify (receiver, e);
  END_PROTECTED
  return ret;
}

void
Application::exit (int result)
{
  shutdown ();
  ::exit (result);
}

void
Application::shutdown ()
{
  if (mp_mw) {
    delete mp_mw;
    mp_mw = 0;
    mp_plugin_root = 0;
  } else if (mp_plugin_root) {
    delete mp_plugin_root;
    mp_plugin_root = 0;
  }

  if (mp_recorder) {
    delete mp_recorder;
    mp_recorder = 0;
  }

  if (mp_ruby_interpreter) {
    delete mp_ruby_interpreter;
    mp_ruby_interpreter = 0;
  }

  //  clean up all static data now, since we don't trust the static destructors
  tl::StaticObjects::cleanup ();

  mp_qapp = 0;
  mp_qapp_gui = 0;
  ms_instance = 0;
}

Application *
Application::instance () 
{
  return ms_instance;
}

std::string
Application::version () const
{
  return std::string (lay::Version::name ()) + " " + lay::Version::version ();
}
      
std::string 
Application::usage () 
{
  std::string r;
  r = std::string (lay::Version::exe_name ()) + " [<options>] [<file>] ..\n";
  r += tl::translate ("options") + "\n";
  r += tl::translate ("  -c <config file>    Use this configuration file") + "\n";
  r += tl::translate ("  -nc                 Don't use a configuration file (implies -t)") + "\n";
  r += tl::translate ("  -d <debug level>    Set debug level") + "\n";
  r += tl::translate ("  -e                  Editable mode (allow editing of files)") + "\n";
  r += tl::translate ("  -ne                 Readonly mode (editing of files is disabled)") + "\n";
  r += tl::translate ("  -gr <file name>     Record GUI test file") + "\n";
  r += tl::translate ("  -gp <file name>     Replay GUI test file") + "\n";
  r += tl::translate ("  -gb <line number>   Replay GUI test file up to (including) line") + "\n";
  r += tl::translate ("  -gx <millisec>      Replay rate for GUI test file") + "\n";
  r += tl::translate ("  -gi                 Incremental logs for GUI test file") + "\n";
  r += tl::translate ("  -i                  Disable undo buffering (less memory requirements)") + "\n";
  r += tl::translate ("  -ni                 Enable undo buffering (default, overrides previous -i option)") + "\n";
  r += tl::translate ("  -j <path>           Add the given path to the macro project paths") + "\n"; 
  r += tl::translate ("  -l <lyp file>       Use layer properties file") + "\n";
  r += tl::translate ("  -lx                 With -l: add other layers as well") + "\n";
  r += tl::translate ("  -lf                 With -l: use the lyp file as it is (no expansion to multiple layouts)") + "\n";
  r += tl::translate ("  -m <database file>  Load RDB (report database) file (into previous layout view)") + "\n";
  r += tl::translate ("  -n <technology>     Technology to use for next layout(s) on command line") + "\n";
  r += tl::translate ("  -p <plugin>         Load the plugin (can be used multiple times)") + "\n";
  if (s_ruby_available) {
    r += tl::translate ("  -r <script>         Execute ruby main script on startup (after having loaded files etc.)") + "\n";
    r += tl::translate ("  -rm <script>        Execute ruby module on startup (can be used multiple times)") + "\n";
    r += tl::translate ("  -rd <name>=<value>  Specify ruby variable") + "\n";
    r += tl::translate ("  -rx                 Ignore all implicit macros (*.rbm, rbainit, *.lym)") + "\n";
  }
  r += tl::translate ("  -s                  Load files into same view") + "\n";
  r += tl::translate ("  -t                  Don't update the configuration file on exit") + "\n";
  r += tl::translate ("  -nt                 Update the configuration file on exit (default, overrides previous -t option)") + "\n";
  r += tl::translate ("  -u <file name>      Restore session from given file") + "\n";
  r += tl::translate ("  -v                  Print program version and exit") + "\n";
  r += tl::translate ("  -wd <name>=<value>  Define a variable within expressions") + "\n";
  r += tl::translate ("  -x                  Synchronous drawing mode") + "\n";
  r += tl::translate ("  -zz                 Non-GUI mode (database only, implies -nc)") + "\n";
  r += tl::translate ("  -z                  Non-GUI mode (hidden views)") + "\n";
  return r;
}

int
Application::run ()
{
  gtf::Player player (0);

  if (mp_mw) {

    mp_mw->set_synchronous (m_sync_mode);

    if (! m_no_gui) {
      mp_mw->setWindowTitle (tl::to_qstring (version ()));
      mp_mw->resize (800, 600);
      mp_mw->show ();
    }

    if (! m_gtf_replay.empty ()) {
      player.load (m_gtf_replay);
    }

    if (mp_recorder) {
      mp_recorder->start ();
    }

  }

  int result = 0;

  bool config_failed = false;

  for (std::vector <std::string>::const_iterator c = m_config_files.begin (); c != m_config_files.end (); ++c) {
    BEGIN_PROTECTED_CLEANUP
      mp_plugin_root->read_config (*c);
      //  if the last config was read successfully no reset will happen:
      config_failed = false;
    END_PROTECTED_CLEANUP {
      config_failed = true;
    }
  }

  if (config_failed) {
    reset_config ();
  }

  if (s_ruby_available) {

    for (std::vector< std::pair<std::string, std::string> >::const_iterator v = m_variables.begin (); v != m_variables.end (); ++v) {
      ruby_interpreter ().define_variable (v->first, v->second);
      tl::log << "Ruby variable definition: " << v->first << "='" << v->second << "'";
    }

    for (std::vector<std::string>::const_iterator m = m_ruby_modules.begin (); m != m_ruby_modules.end (); ++m) {
      BEGIN_PROTECTED 
        ruby_interpreter ().load_file (*m);
        tl::log << "Ruby module loaded '" << *m << "'";
      END_PROTECTED
    }

  }

  //  scan for libraries
  for (std::vector <std::string>::const_iterator p = m_klayout_path.begin (); p != m_klayout_path.end (); ++p) {

    QDir lp = QDir (tl::to_qstring (*p)).filePath (tl::to_qstring ("libraries"));

    QStringList name_filters;
    name_filters << QString::fromAscii ("*");

    QStringList libs = lp.entryList (name_filters, QDir::Files);
    for (QStringList::const_iterator im = libs.begin (); im != libs.end (); ++im) {

      std::string filename = tl::to_string (*im);

      try {

        std::auto_ptr<db::Library> lib (new db::Library ());
        lib->set_description (filename);
        lib->set_name (tl::to_string (QFileInfo (*im).baseName ()));

        tl::log << "Reading library '" << filename << "'";
        tl::InputZLibFile stream (tl::to_string (lp.filePath (*im)));
        db::Reader reader (stream);
        reader.read (lib->layout ());

        //  Use the libname if there is one
        for (db::Reader::meta_info_iterator m = reader.begin_meta (); m != reader.end_meta (); ++m) {
            if (m->name == "libname" && ! m->value.empty ()) {
                lib->set_name (m->value);
                break;
            }
        }

        db::LibraryManager::instance ().register_lib (lib.release ());

      } catch (tl::Exception &ex) {
        tl::error << ex.msg ();
      }

    }

  }

  //  run all autorun macros
  lay::MacroCollection::root ().autorun ();

  if (mp_mw) {

    for (std::vector <std::pair<file_type, std::pair<std::string, std::string> > >::const_iterator f = m_files.begin (); f != m_files.end (); ++f) {

      if (f->first == layout_file || f->first == layout_file_with_tech) {

        std::string filename = f->second.first;

        if (f->first != layout_file_with_tech) {
          mp_mw->add_mru (f->second.first);
          mp_mw->load_layout (f->second.first, m_same_view ? 2 /*same view*/ : 1 /*new view*/);
        } else {
          mp_mw->add_mru (f->second.first, f->second.second);
          mp_mw->load_layout (f->second.first, f->second.second, m_same_view ? 2 /*same view*/ : 1 /*new view*/);
        }

        //  Make the first one loaded the active one.
        if (mp_mw->current_view ()) {
          mp_mw->current_view ()->set_active_cellview_index (0);
        }

      } else {

        if (mp_mw->current_view () == 0) {
          mp_mw->create_view ();
        }

        if (mp_mw->current_view () != 0) {
          std::auto_ptr <rdb::Database> db (new rdb::Database ());
          db->load (f->second.first);
          int rdb_index = mp_mw->current_view ()->add_rdb (db.release ());
          mp_mw->current_view ()->open_rdb_browser (rdb_index, mp_mw->current_view ()->active_cellview_index ());
        }

      }
    }

    if (! m_layer_props_file.empty ()) {

      if (m_lyp_map_all_cvs && mp_mw->is_single_cv_layer_properties_file (m_layer_props_file)) {
        mp_mw->load_layer_properties (m_layer_props_file, -1, true /*all views*/, m_lyp_add_default);
      } else {
        mp_mw->load_layer_properties (m_layer_props_file, true /*all views*/, m_lyp_add_default);
      }

      tl::log << "Layer properties loaded '" << m_layer_props_file << "'";

      //  because the layer may carry transformations, we need to refit the cellviews.
      for (unsigned int v = 0; v != mp_mw->views (); ++v) {
        mp_mw->view (v)->zoom_fit ();
      }

    }

    if (! m_session_file.empty ()) {
      mp_mw->restore_session (m_session_file);
      tl::log << "Session restored '" << m_session_file << "'";
    }

    if (! m_gtf_replay.empty ()) {
      player.replay (m_gtf_replay_rate, m_gtf_replay_stop);
    }

  }

  if (! m_ruby_script.empty () && s_ruby_available) {
    ruby_interpreter ().load_file (m_ruby_script);
    //  TODO: allow the ruby interpreter to return some value 
    result = 0;
  } else {

    if (! m_no_gui && mp_mw && m_gtf_replay.empty () && ! mp_recorder) {
      //  Show initial tip window if required
      mp_mw->about_to_exec ();
    }

    result = exec ();
  }

  //  save the recorded test events
  if (mp_mw && mp_recorder && mp_recorder->recording ()) {
    mp_recorder->stop ();
    mp_recorder->save ();
  }

  if (m_write_config_file) {
    if (! m_config_file_to_write.empty ()) {
      mp_plugin_root->write_config (m_config_file_to_write); 
    }
    if (! m_config_file_to_delete.empty () && m_config_file_to_delete != m_config_file_to_write) {
      QFile::remove (tl::to_qstring (m_config_file_to_delete));
    }
  }

  return result;
}

static void 
dump_children (QObject *obj, int level = 0)
{
  QObjectList children = obj->children ();
  if (! children.isEmpty () || ! obj->objectName ().isEmpty ()) {
    std::string info;
    for (int i = 0; i < level; ++i) {
      info += "  ";
    }
    if (obj->objectName ().isEmpty ()) {
      info += "<unnamed>";
    } else {
      info += tl::to_string (obj->objectName ());
    }
    tl::info << info;
    for (QObjectList::const_iterator child = children.begin (); child != children.end (); ++child) {
      dump_children (*child, level + 1);
    }
  }
}

int 
Application::exec ()
{
  if (m_no_gui) {
    return 0;
  } else {

    //  if requested, dump the widgets
    if (tl::verbosity () >= 40) {

      QWidgetList tl_widgets = QApplication::topLevelWidgets (); 

      tl::info << tl::translate ("Widget tree:");
      for (QWidgetList::const_iterator tl = tl_widgets.begin (); tl != tl_widgets.end (); ++tl) { 
        if (! (*tl)->objectName ().isEmpty ()) {
          dump_children (*tl); 
        }
      }
      tl::info << "";

      tl::info << tl::translate ("Actions list:");
      for (QWidgetList::const_iterator tl = tl_widgets.begin (); tl != tl_widgets.end (); ++tl) { 
        if (! (*tl)->objectName ().isEmpty ()) {
          QList<QAction *> actions = (*tl)->findChildren<QAction *> ();
          if (! actions.isEmpty ()) {
            tl::info << tl::to_string ((*tl)->objectName ()) << ":";
            for (QList<QAction *>::const_iterator a = actions.begin (); a != actions.end (); ++a) {
              if (! (*a)->objectName ().isEmpty ()) {
                tl::info << "  " << tl::to_string ((*a)->objectName ());
              }
            }
          }
        }
      }
      tl::info << "";

    }

    return mp_qapp_gui->exec ();

  }
}

void
Application::process_events ()
{
  if (! m_no_gui) {
    QApplication::processEvents ();
  }
}

bool 
Application::write_config (const std::string &config_file)
{
  return mp_plugin_root ? mp_plugin_root->write_config (config_file) : 0;
}

void 
Application::reset_config ()
{
  clear_config ();
  for (std::vector <std::string>::const_iterator c = m_initial_config_files.begin (); c != m_initial_config_files.end (); ++c) {
    try {
      read_config (*c);
    } catch (...) { }
  }
}

void 
Application::clear_config ()
{
  if (mp_plugin_root) {
    mp_plugin_root->clear_config ();
  }
}

bool 
Application::read_config (const std::string &config_file)
{
  return mp_plugin_root ? mp_plugin_root->read_config (config_file) : true;
}

void 
Application::set_config (const std::string &name, const std::string &value)
{
  if (mp_plugin_root) {
    mp_plugin_root->config_set (name, value);
  }
}

std::string 
Application::get_config (const std::string &name) const
{
  if (mp_plugin_root) {
    return mp_plugin_root->config_get (name);
  } else {
    return std::string ();
  }
}

std::vector<std::string> 
Application::get_config_names () const
{
  std::vector<std::string> names;
  if (mp_plugin_root) {
    mp_plugin_root->get_config_names (names);
  }
  return names;
}

}

