
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2013 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layMacro.h"
#include "layMainWindow.h"
#include "layAbstractMenu.h"
#include "layApplication.h"
#include "layExceptions.h"
#include "rba.h"
#include "gsiDecl.h"

#include "tlString.h"
#include "tlLog.h"
#include "tlXMLParser.h"

#include <QtCore/QFile>
#include <QtCore/QDir>
#include <QtCore/QUrl>
#include <QtCore/QResource>

#include <fstream>

namespace lay
{

// ----------------------------------------------------------------------

Macro::Macro (Format format)
  : m_modified (true), m_readonly (false), m_autorun (false), m_autorun_early (false), m_show_in_menu (false), m_is_file (false), mp_parent (0), m_interpreter (None), m_format (format)
{
  // .. nothing yet ..
}

void Macro::on_changed ()
{
  emit changed ();
  if (mp_parent) {
    mp_parent->on_macro_changed (this);
  }
}

void Macro::assign (const lay::Macro &other)
{
  m_description = other.m_description;
  m_version = other.m_version;
  m_prolog = other.m_prolog;
  m_epilog = other.m_epilog;
  m_text = other.m_text;
  m_doc = other.m_doc;
  m_version = other.m_version;
  m_needs_load_from = other.m_needs_load_from;
  m_modified = other.m_modified;
  m_readonly = other.m_readonly;
  m_autorun = other.m_autorun;
  m_autorun_early = other.m_autorun_early;
  m_show_in_menu = other.m_show_in_menu;
  m_shortcut = other.m_shortcut;
  m_interpreter = other.m_interpreter;
  m_format = other.m_format;
  m_group_name = other.m_group_name;
  m_menu_path = other.m_menu_path;
  m_format = other.m_format;
  m_interpreter = other.m_interpreter;
  m_is_file = other.m_is_file;
  on_changed ();
}

bool Macro::operator== (const Macro &other) const
{
  return 
    m_description == other.m_description &&
    m_version == other.m_version &&
    m_epilog == other.m_epilog &&
    m_prolog == other.m_prolog &&
    m_text == other.m_text &&
    m_needs_load_from == other.m_needs_load_from &&
    m_autorun == other.m_autorun &&
    m_autorun_early == other.m_autorun_early &&
    m_show_in_menu == other.m_show_in_menu &&
    m_shortcut == other.m_shortcut &&
    m_interpreter == other.m_interpreter &&
    m_format == other.m_format;
}

void Macro::save ()
{
  save_to (path ());
}

bool Macro::del ()
{
  if (m_is_file) {
    if (tl::verbosity () >= 20) {
      tl::log << "Deleting macro " << path ();
    }
    QFile f (tl::to_qstring (path ()));
    return f.remove ();
  } else {
    return true;
  }
}

struct Interpreter2s
{
  std::string to_string (Macro::Interpreter i)
  {
    switch (i) {
      case Macro::Ruby:
        return "ruby";
      default:
        return "none";
    }
  }

  void from_string (const std::string &s, Macro::Interpreter &i)
  {
    if (s == "ruby") {
      i = Macro::Ruby;
    } else {
      i = Macro::None;
    }
  }
};

struct Format2s
{
  std::string to_string (Macro::Format f)
  {
    switch (f) {
      case Macro::MacroFormat:
        return "general";
      case Macro::RubyFormat:
        return "ruby";
      case Macro::TextFormat:
        return "text";
      default:
        return "none";
    }
  }

  void from_string (const std::string &s, Macro::Format &f)
  {
    if (s == "general") {
      f = Macro::MacroFormat;
    } else if (s == "ruby") {
      f = Macro::RubyFormat;
    } else if (s == "text") {
      f = Macro::TextFormat;
    } else {
      f = Macro::NoFormat;
    }
  }
};

/**
 *  @brief Declaration of the XML structure of a macro
 */
static tl::XMLStruct<lay::Macro> xml_struct ("klayout-macro", 
  tl::make_member (&Macro::description, &Macro::set_description, "description") +
  tl::make_member (&Macro::version, &Macro::set_version, "version") +
  tl::make_member (&Macro::prolog, &Macro::set_prolog, "prolog") +
  tl::make_member (&Macro::epilog, &Macro::set_epilog, "epilog") +
  tl::make_member (&Macro::doc, &Macro::set_doc, "doc") +
  tl::make_member (&Macro::format, &Macro::set_format, "format", Format2s ()) +
  tl::make_member (&Macro::is_autorun, &Macro::set_autorun, "autorun") +
  tl::make_member (&Macro::is_autorun_early, &Macro::set_autorun_early, "autorun-early") +
  tl::make_member (&Macro::shortcut, &Macro::set_shortcut, "shortcut") +
  tl::make_member (&Macro::show_in_menu, &Macro::set_show_in_menu, "show-in-menu") +
  tl::make_member (&Macro::group_name, &Macro::set_group_name, "group-name") +
  tl::make_member (&Macro::menu_path, &Macro::set_menu_path, "menu-path") +
  tl::make_member (&Macro::interpreter, &Macro::set_interpreter, "interpreter", Interpreter2s ()) +
  tl::make_member (&Macro::text, &Macro::set_text, "text") 
);

void Macro::save_to (const std::string &path)
{
  tl::OutputFile of (path);
  if (m_format == MacroFormat) {
    if (tl::verbosity () >= 20) {
      tl::log << "Saving macro to " << path;
    }
    tl::OutputStream os (of);
    xml_struct.write (os, *this);
  } else if (m_format == RubyFormat || m_format == TextFormat) {
    tl::OutputStream os (of);
    os << text ();
  }
  if (m_modified || ! m_is_file) {
    m_modified = false;
    m_is_file = true;
    on_changed ();
  }
}

Macro::Format Macro::load_from (const std::string &path)
{
  Format format = NoFormat;

  if (tl::verbosity () >= 20) {
    tl::log << "Loading macro from " << path;
  }

  std::string suffix = tl::to_string (QFileInfo (tl::to_qstring (path)).suffix ());
  if (suffix == "lym") {

    format = MacroFormat;

    tl::XMLFileSource source (path);
    xml_struct.parse (source, *this);

  } else if (suffix == "rb" || suffix == "txt") {

    format = (suffix == "rb" ? RubyFormat : TextFormat);

    //  Delay-load the file
    m_text.clear ();
    m_needs_load_from = path;

    if (format == RubyFormat) {
      m_interpreter = Ruby;
    } else {
      m_interpreter = None;
    }

  } else {
    throw tl::Exception (tl::translate ("Unable to determine format for file from suffix ") + path);
  }

  m_modified = true;
  on_changed ();

  return format;
}

Macro::Format Macro::load_from_string (const std::string &text, const std::string &url)
{
  Format format = NoFormat;

  if (tl::verbosity () >= 20) {
    tl::log << "Loading macro from " << url;
  }

  std::string suffix = tl::to_string (QFileInfo (QUrl (tl::to_qstring (url)).path ()).suffix ());
  if (suffix == "lym") {

    format = MacroFormat;

    tl::XMLStringSource source (text);
    xml_struct.parse (source, *this);

  } else if (suffix == "rb" || suffix == "txt") {

    format = (suffix == "rb" ? RubyFormat : TextFormat);

    m_text = text;

    if (format == RubyFormat) {
      m_interpreter = Ruby;
    } else {
      m_interpreter = None;
    }

  } else {
    throw tl::Exception (tl::translate ("Unable to determine format for file from suffix " + url));
  }

  m_modified = true;
  on_changed ();

  return format;
}

void Macro::load ()
{
  m_format = load_from (path ());
  m_modified = false;
  m_is_file = true;
  on_changed ();
}

Macro::Format Macro::format_from_suffix (const std::string &fn)
{
  std::string suffix = tl::to_string (QFileInfo (tl::to_qstring (fn)).suffix ());

  if (suffix == "rb") {
    return Macro::RubyFormat;
  } else if (suffix == "lym") {
    return Macro::MacroFormat;
  } else if (suffix == "txt") {
    return Macro::TextFormat;
  } else {
    return Macro::NoFormat;
  }
}

std::string Macro::suffix_for_format (Macro::Format format)
{
  if (format == Macro::RubyFormat) {
    return ".rb";
  } else if (format == Macro::MacroFormat) {
    return ".lym";
  } else if (format == Macro::TextFormat) {
    return ".txt";
  } else {
    return std::string ();
  }
}

std::string Macro::path () const
{
  std::string suffix = suffix_for_format (m_format);
  if (mp_parent) {
    return tl::to_string (QFileInfo (QDir (tl::to_qstring (mp_parent->path ())), tl::to_qstring (m_name + suffix)).filePath ());
  } else {
    return m_name + suffix;
  }
}

void Macro::set_is_file ()
{
  if (! m_is_file) {
    m_is_file = true;
    on_changed ();
  }
}

void Macro::reset_modified ()
{
  if (m_modified) {
    m_modified = false;
    on_changed ();
  }
}

bool Macro::rename (const std::string &n)
{
  if (m_is_file) {
    std::string suffix = suffix_for_format (m_format);
    if (tl::verbosity () >= 20) {
      tl::log << "Renaming macro " << path () << " to " << n;
    }
    QFile f (tl::to_qstring (path ()));
    if (! f.rename (QFileInfo (QDir (tl::to_qstring (mp_parent->path ())), tl::to_qstring (n + suffix)).filePath ())) {
      return false;
    }
  }

  if (parent ()) {
    parent ()->rename_macro (this, n);
  }

  m_name = n;
  on_changed ();
  return true;
}

std::string Macro::dir () const
{
  if (mp_parent) {
    return mp_parent->path ();
  } else {
    return std::string ();
  }
}

std::string Macro::display_string () const
{
  std::string r = name ();
  if (! m_description.empty ()) {
    r += " - " + m_description;
  }
  if (! m_shortcut.empty ()) {
    r += " (" + m_shortcut + ")";
  }
  return r;
}

void Macro::set_doc (const std::string &d)
{
  if (m_doc != d) {
    m_modified = true;
    m_doc = d;
    on_changed ();
  }
}

void Macro::set_description (const std::string &d)
{
  if (m_description != d) {
    m_modified = true;
    m_description = d;
    on_changed ();
  }
}

void Macro::set_epilog (const std::string &s)
{
  if (m_epilog != s) {
    m_modified = true;
    m_epilog = s;
    on_changed ();
  }
}

void Macro::set_prolog (const std::string &s)
{
  if (m_prolog != s) {
    m_modified = true;
    m_prolog = s;
    on_changed ();
  }
}

void Macro::set_version (const std::string &s)
{
  if (m_version != s) {
    m_modified = true;
    m_version = s;
    on_changed ();
  }
}

const std::string &Macro::text () const
{
  if (! m_needs_load_from.empty ()) {

    //  Read the whole input file
    tl::InputFile file (m_needs_load_from);
    m_text.clear ();
    char b[1024];
    size_t n = 0;
    while ((n = file.read (b, sizeof (b))) > 0) {
      m_text += std::string (b, 0, n);
    }

    m_needs_load_from.clear ();

  }

  return m_text;
}

void Macro::set_text (const std::string &t)
{
  if (text () != t) {
    m_modified = true;
    m_text = t;
    on_changed ();
  }
}

void Macro::set_format (Format f)
{
  if (f != m_format) {
    m_modified = true;
    m_format = f;
    on_changed ();
  }
}

void Macro::set_interpreter (Interpreter i)
{
  if (i != m_interpreter) {
    m_modified = true;
    m_interpreter = i;
    on_changed ();
  }
}

void Macro::set_autorun_early (bool f)
{
  if (f != m_autorun_early) {
    m_modified = true;
    m_autorun_early = f;
    on_changed ();
  }
}

void Macro::set_autorun (bool f)
{
  if (f != m_autorun) {
    m_modified = true;
    m_autorun = f;
    on_changed ();
  }
}

void Macro::set_show_in_menu (bool f)
{
  if (f != m_show_in_menu) {
    m_modified = true;
    m_show_in_menu = f;
    on_changed ();
  }
}

void Macro::set_menu_path (const std::string &mp)
{
  if (m_menu_path != mp) {
    m_modified = true;
    m_menu_path = mp;
    on_changed ();
  }
}

void Macro::set_group_name (const std::string &g)
{
  if (m_group_name != g) {
    m_modified = true;
    m_group_name = g;
    on_changed ();
  }
}

void Macro::set_shortcut (const std::string &s)
{
  if (s != m_shortcut) {
    m_modified = true;
    m_shortcut = s;
    on_changed ();
  }
}

void Macro::set_readonly (bool f)
{
  if (m_readonly != f) {
    m_readonly = f;
    on_changed ();
  }
}

class ExternalMethod
  : public gsi::MethodBase
{
public:
  ExternalMethod (const std::string &name, const std::string &doc, bool c, bool s)
    : gsi::MethodBase (name, doc, c, s)
  {
    //  no return type
    gsi::ArgType a;
    a.type = gsi::BasicType (-1);
    set_return (a);
  }

  virtual MethodBase *clone () const 
  {
    return new ExternalMethod (*this);
  }

  //  this class is not intended to go functional. It's just a hook for the documentation
  virtual void call(void*, gsi::SerialArgs&, gsi::SerialArgs&) const
  {
    tl_assert (false); 
  }
};

class ExternalClass 
  : public gsi::ClassBase
{
public:
  ExternalClass (const std::string &name, const gsi::ClassBase *base, const std::string &doc, const gsi::Methods &mm)
    : gsi::ClassBase (0, name, base, doc, mm)
  {
    //  .. nothing yet ..
  }

  //  this class is not intended to go functional. It's just a hook for the documentation
  gsi::ClassBase *declaration () const { return 0; }
  void *client_data (void *p) const { tl_assert (false); }
  virtual void destroy (void *p) const { tl_assert (false); }
  virtual void *create () const { tl_assert (false); }
  virtual void assign (void *dest, void *src) const { tl_assert (false); }
  virtual void get_vector_of (gsi::SerialArgs &from, const gsi::ArgType &a, void *data, void (*cb) (void *data, void *obj)) const { tl_assert (false); } 
  void push_vector_of (gsi::SerialArgs &to, const gsi::ArgType &a, std::list<gsi::TempObject> &temp_objects, const std::vector<void *> &objects) const { tl_assert (false); } 
  virtual void *take_and_create (gsi::SerialArgs &from) const { tl_assert (false); }
  virtual void copy_to (gsi::SerialArgs &at, void *from) const { tl_assert (false); } 
  virtual void *create_obj_iterator (gsi::SerialArgs &from) const { tl_assert (false); }
  virtual bool obj_iterator_at_end (void *iter) const { tl_assert (false); }
  virtual void *take_iterator_obj (void *iter) const { tl_assert (false); }
  virtual const void *get_iterator_obj (void *iter) const { tl_assert (false); }
  virtual void inc_obj_iterator (void *iter) const { tl_assert (false); }
  virtual void delete_obj_iterator (void *iter) const { tl_assert (false); }
  virtual bool can_destroy () const { return false; }
  virtual bool can_copy () const { return false; }
  virtual bool can_default_create () const { return false; }
  virtual const gsi::ClassBase *subclass_decl (const void *p) const { tl_assert (false); }
  virtual const gsi::ClassBase *find_subclass_decl (const std::type_info &ti) const  { tl_assert (false); }
};

void Macro::install_doc () const
{
  std::vector<std::string> lines = tl::split (tl::trim (doc ()), "\n");
  if (! lines.empty () && tl::trim (lines [0]).find ("@class") == 0) {

    //  this macro provides documentation for the GSI namespace
    gsi::ClassBase *cls = 0;

    for (size_t i = 0; i < lines.size (); ++i) {

      bool st = false;
      tl::Extractor ex (lines [i].c_str ());
      if (ex.test ("@class")) {

        std::string cls_name, super_cls_name;
        ex.read_word_or_quoted (cls_name);
        if (ex.test ("<")) {
          ex.read_word_or_quoted (super_cls_name);
        }

        std::string doc;
        while (++i < lines.size ()) {
          std::string l = tl::trim (lines [i]);
          if (l.find ("@method") == 0 || l.find ("@static_method") == 0) {
            break;
          }
          if (! doc.empty ()) {
            doc += "\n";
          }
          doc += lines [i];
        }
        --i;

        if (cls) {
          tl::error << tl::translate ("Reading class doc from ") << path () << ": " << tl::translate ("Duplicate @class");
          return;
        }

        const gsi::ClassBase *super_cls = 0;
        for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
          if (c->name () == cls_name) {
            cls = const_cast <gsi::ClassBase *> (&*c);
          } else if (c->name () == super_cls_name) {
            super_cls = &*c;
          }
        }

        if (! super_cls && ! super_cls_name.empty ()) {
          tl::error << tl::translate ("Reading class doc from ") << path () << ": " << tl::translate ("Cannot find super class: ") << super_cls_name;
          return;
        }

        if (! cls) {
          //  create a new class declaration
          static tl::stable_vector<ExternalClass> ext_classes;
          ExternalClass *ext_cls = new ExternalClass (cls_name, super_cls, doc, gsi::Methods ()); 
          ext_classes.push_back (ext_cls);
          cls = ext_cls;
        }

      } else if (ex.test ("@method") || (st = ex.test ("@static_method")) == true) {

        tl_assert (cls != 0);

        std::string n;
        ex.read_word_or_quoted (n);

        std::string doc;
        while (++i < lines.size ()) {
          std::string l = tl::trim (lines [i]);
          if (l.find ("@method") == 0 || l.find ("@static_method") == 0) {
            break;
          }
          if (! doc.empty ()) {
            doc += "\n";
          }
          doc += lines [i];
        }
        --i;

        ExternalMethod *meth = new ExternalMethod (n, doc, false, st);
        cls->add_method (meth);

      }

    }

  }
}

void Macro::run () const
{
  if (tl::verbosity () >= 20) {
    tl::log << "Running macro " << path ();
  }

  try {
    if (interpreter () == lay::Macro::Ruby) {
      if (! prolog ().empty ()) {
        lay::Application::instance ()->ruby_interpreter ().eval_string (prolog ().c_str ());
      }
      lay::Application::instance ()->ruby_interpreter ().eval_string (text ().c_str (), path ().c_str (), 1);
      if (! epilog ().empty ()) {
        lay::Application::instance ()->ruby_interpreter ().eval_string (epilog ().c_str ());
      }
    }
  } catch (rba::ExitException &) {
    //  .. ignore exit exceptions ..
  }
}

// ----------------------------------------------------------------------

static MacroCollection ms_root;

MacroCollection::MacroCollection ()
  : mp_parent (0), m_virtual (true), m_readonly (false)
{
  // .. nothing yet ..
}

MacroCollection::~MacroCollection ()
{
  for (iterator m = begin (); m != end (); ++m) {
    delete m->second;
  }
  m_macros.clear ();

  for (child_iterator mm = begin_children (); mm != end_children (); ++mm) {
    delete mm->second;
  }
  m_folders.clear ();
}

void MacroCollection::begin_changes ()
{
  //  Note: it is very important that each on_changed occurs after exactly one begin_changes.
  //  (See #459 for example)
  if (mp_parent) {
    mp_parent->begin_changes ();
  } else {
    emit about_to_change ();
  }
}

void MacroCollection::on_changed ()
{
  //  Note: it is very important that each on_changed occurs after exactly one begin_changes.
  //  (See #459 for example)
  emit changed ();
  on_macro_collection_changed (this);
}

void MacroCollection::on_macro_collection_changed (MacroCollection *mc)
{
  if (mp_parent) {
    mp_parent->on_macro_collection_changed (mc);
  } else {
    emit macro_collection_changed (mc);
  }
}

void MacroCollection::on_child_deleted (MacroCollection *mc)
{
  emit child_deleted (mc);
  on_macro_collection_deleted (mc);
}

void MacroCollection::on_macro_collection_deleted (MacroCollection *mc)
{
  if (mp_parent) {
    mp_parent->on_macro_collection_deleted (mc);
  } else {
    emit macro_collection_deleted (mc);
  }
}

void MacroCollection::on_macro_deleted_here (Macro *macro)
{
  emit macro_deleted_here (macro);
  on_macro_deleted (macro);
}

void MacroCollection::on_macro_deleted (Macro *macro)
{
  if (mp_parent) {
    mp_parent->on_macro_deleted (macro);
  } else {
    emit macro_deleted (macro);
  }
}

void MacroCollection::on_macro_changed (Macro *macro)
{
  if (mp_parent) {
    mp_parent->on_macro_changed (macro);
  } else {
    emit macro_changed (macro);
  }
}

void MacroCollection::collect_used_nodes(std::set <Macro *> &macros, std::set <MacroCollection *> &macro_collections)
{
  for (MacroCollection::child_iterator c = begin_children (); c != end_children (); ++c) {
    macro_collections.insert (c->second);
    c->second->collect_used_nodes (macros, macro_collections);
  }
  for (MacroCollection::iterator c = begin (); c != end (); ++c) {
    macros.insert (c->second);
  }
}

Macro *MacroCollection::macro_by_name (const std::string &name, Macro::Format format)
{
  std::multimap <std::string, Macro *>::iterator i = m_macros.find (name);
  while (i != m_macros.end () && i->first == name) {
    if (format == Macro::NoFormat || i->second->format () == format) {
      return i->second;
    }
    ++i;
  } 
  return 0;
}

const Macro *MacroCollection::macro_by_name (const std::string &name, Macro::Format format) const
{
  std::multimap <std::string, Macro *>::const_iterator i = m_macros.find (name);
  while (i != m_macros.end () && i->first == name) {
    if (format == Macro::NoFormat || i->second->format () == format) {
      return i->second;
    }
    ++i;
  } 
  return 0;
}

MacroCollection *MacroCollection::folder_by_name (const std::string &name)
{
  std::map <std::string, MacroCollection *>::iterator i = m_folders.find (name);
  if (i != m_folders.end ()) {
    return i->second;
  } else {
    return 0;
  }
}

const MacroCollection *MacroCollection::folder_by_name (const std::string &name) const
{
  std::map <std::string, MacroCollection *>::const_iterator i = m_folders.find (name);
  if (i != m_folders.end ()) {
    return i->second;
  } else {
    return 0;
  }
}

std::string MacroCollection::path () const
{
  if (m_virtual) {
    return m_path;
  } else if (mp_parent) {
    return tl::to_string (QFileInfo (QDir (tl::to_qstring (mp_parent->path ())), tl::to_qstring (m_path)).filePath ());
  } else {
    return m_path;
  }
}

std::string MacroCollection::display_string () const
{
  if (m_virtual) {
    return "[" + m_description + "]";
  } else {
    std::string r = name ();
    if (! m_description.empty ()) {
      r += " - " + m_description;
    }
    return r;
  }
}

void
MacroCollection::make_readonly (bool f)
{
  if (m_readonly != f) {
    begin_changes ();
    m_readonly = f;
    on_changed ();
  }
}

MacroCollection *
MacroCollection::add_folder (const std::string &description, const std::string &path, bool readonly)
{
  if (! path.empty () && path[0] == ':') {
    readonly = true;
  } else {

    QFileInfo file_info (tl::to_qstring (path));

    if (! file_info.exists ()) {

      //  Try to create the folder since it does not exist yet
      if (tl::verbosity () >= 20) {
        tl::log << "Folder does not exist yet - trying to create it: " << path;
      }
      if (! QDir::root ().mkpath (file_info.absoluteFilePath ())) {
        if (tl::verbosity () >= 10) {
          tl::error << "Unable to create folder path: " << path;
        }
        return 0;
      }

      file_info.refresh ();

    }

    if (! file_info.isDir ()) {
      if (tl::verbosity () >= 10) {
        tl::error << "Folder is not a directory: " << path;
      }
      return 0;
    }

    QString cp = file_info.canonicalFilePath ();
    if (cp.isEmpty ()) {
      return 0;
    }

    for (child_iterator f = m_folders.begin (); f != m_folders.end (); ++f) {
      //  skip, if that folder is in the collection already
      if (QFileInfo (tl::to_qstring (f->first)).canonicalFilePath () == cp) {
        return 0;
      }
    }

    if (! readonly && ! file_info.isWritable ()) {
      readonly = true;
      if (tl::verbosity () >= 20) {
        tl::log << "Folder is read-only: " << path;
      }
    }

  }

  begin_changes ();

  MacroCollection *mc = m_folders.insert (std::make_pair (path, new MacroCollection ())).first->second;
  mc->set_parent (this);
  mc->set_name (path);
  mc->set_description (description);
  mc->set_readonly (readonly);
  mc->scan (path, readonly);

  on_changed ();

  return mc;
}

void MacroCollection::scan (const std::string &path, bool readonly)
  {
    if (tl::verbosity () >= 20) {
      tl::info << "Scanning macro path " << path << " (readonly=" << readonly << ")";
    }

    if (! path.empty () && path[0] == ':') {

      //  look for an index file
      QResource res (tl::to_qstring (path + "/index.txt"));
      QByteArray data;
      if (res.isCompressed ()) {
        data = qUncompress ((const unsigned char *)res.data (), (int)res.size ());
      } else {
        data = QByteArray ((const char *)res.data (), (int)res.size ());
      }

      //  Read index file
      std::vector<std::string> lines = tl::split (std::string (data.constData (), data.size ()), "\n");
      std::string description_prefix;
      for (std::vector<std::string>::const_iterator l = lines.begin (); l != lines.end (); ++l) {

        std::string ll = tl::trim (*l);
        if (! ll.empty () && ll [0] != '#') {

          std::string url = path + "/" + ll;
          QResource res (tl::to_qstring (url));
          if (res.size () > 0) {

            QByteArray data;
            if (res.isCompressed ()) {
              data = qUncompress ((const unsigned char *)res.data (), (int)res.size ());
            } else {
              data = QByteArray ((const char *)res.data (), (int)res.size ());
            }

            try {

              std::string n = tl::to_string (QFileInfo (tl::to_qstring (ll)).baseName ());
              Macro *m = m_macros.insert (std::make_pair (n, new Macro (Macro::MacroFormat)))->second;
              m->set_parent (this);
              m->set_name (n);
              m->load_from_string (std::string (data.constData (), data.size ()), url);
              m->set_readonly (readonly);
              m->reset_modified ();
              m->set_is_file ();

            } catch (tl::Exception &ex) {
              tl::error << "Reading " << url << ": " << ex.msg ();
            }

          } else {
            tl::error << "te resource " << url << " not found";
          }

        }

      }

    } else {

      QDir dir (tl::to_qstring (path));
      QStringList filters;
      filters << QString::fromAscii ("*.lym");
      filters << QString::fromAscii ("*.rb");
      filters << QString::fromAscii ("*.txt");

      QStringList files = dir.entryList (filters, QDir::Files);
      for (QStringList::ConstIterator f = files.begin (); f != files.end (); ++f) {

        try {

          std::string n = tl::to_string (QFileInfo (*f).baseName ());
          std::string suffix = tl::to_string (QFileInfo (*f).suffix ());

          Macro::Format format = Macro::NoFormat;
          if (suffix == "rb") {
            format = Macro::RubyFormat;
          } else if (suffix == "lym") {
            format = Macro::MacroFormat;
          } else if (suffix == "txt") {
            format = Macro::TextFormat;
          }

          if (format != Macro::NoFormat) {
            Macro *m = m_macros.insert (std::make_pair (n, new Macro (format)))->second;
            m->set_parent (this);
            m->set_name (n);
            m->load ();
            m->set_readonly (readonly);
          }

        } catch (tl::Exception &ex) {
          tl::error << ex.msg ();
        }

      }

      QStringList folders = dir.entryList (QDir::Dirs | QDir::NoDotAndDotDot);
      for (QStringList::ConstIterator f = folders.begin (); f != folders.end (); ++f) {

        try {

          std::string n = tl::to_string (*f);
          MacroCollection *mc = m_folders.insert (std::make_pair (n, new MacroCollection ())).first->second;
          mc->set_parent (this);
          mc->set_name (n);
          mc->set_virtual (false);
          bool ro = (readonly || ! QFileInfo (dir.filePath (*f)).isWritable ());
          mc->scan (tl::to_string (dir.filePath (*f)), ro);
          mc->set_readonly (ro);

        } catch (tl::Exception &ex) {
          tl::error << ex.msg ();
        }

      }

    }
  }

void MacroCollection::erase (lay::Macro *mp)
{
  for (iterator m = m_macros.begin (); m != m_macros.end (); ++m) {
    if (m->second == mp) {
      begin_changes ();
      on_macro_deleted_here (mp);
      m_macros.erase (m);
      delete mp;
      on_changed ();
      return;
    }
  }
}

void MacroCollection::erase (lay::MacroCollection *mp)
{
  for (child_iterator f = m_folders.begin (); f != m_folders.end (); ++f) {
    if (f->second == mp) {
      begin_changes ();
      on_child_deleted (mp);
      m_folders.erase (f);
      delete mp;
      on_changed ();
      return;
    }
  }
}

void MacroCollection::erase (iterator i) 
{
  begin_changes ();
  on_macro_deleted_here (i->second);
  delete i->second;
  m_macros.erase (i);
  on_changed ();
}

void MacroCollection::erase (child_iterator i)
{
  begin_changes ();
  on_child_deleted (i->second);
  delete i->second;
  m_folders.erase (i);
  on_changed ();
}

void MacroCollection::save ()
{
  for (child_iterator f = m_folders.begin (); f != m_folders.end (); ++f) {
    f->second->save ();
  }

  for (iterator m = m_macros.begin (); m != m_macros.end (); ++m) {
    if (m->second->is_modified () && ! m->second->is_readonly () && ! m->second->path ().empty ()) {
      try {
        m->second->save ();
      } catch (tl::Exception &ex) {
        tl::error << ex.msg ();
      }
    }
  }
}

bool MacroCollection::rename (const std::string &n)
{
  if (tl::verbosity () >= 20) {
    tl::info << "Renaming macro folder " << path () << " to " << n;
  }
  QFile f (tl::to_qstring (path ()));
  begin_changes ();
  if (! f.rename (QFileInfo (QDir (tl::to_qstring (mp_parent->path ())), tl::to_qstring (n)).filePath ())) {
    on_changed ();
    return false;
  } else {
    m_path = n;
    on_changed ();
    return true;
  }
}

lay::MacroCollection *MacroCollection::create_folder (const char *prefix, bool mkdir)
{
  std::string name;
  int n = 0;
  do {
    name = (prefix ? prefix : "new_folder");
    if (n > 0) {
      name += "_" + tl::to_string (n);
    }
    if (m_folders.find (name) == m_folders.end ()) {
      break;
    }
    ++n;
  } while (true);

  if (mkdir && ! QDir (tl::to_qstring (path ())).mkdir (tl::to_qstring (name))) {
    return 0;
  }

  begin_changes ();

  lay::MacroCollection *m = m_folders.insert (std::make_pair (name, new lay::MacroCollection ())).first->second; 
  m->set_virtual (false);
  m->set_name (name);
  m->set_parent (this);

  on_changed ();

  return m;
}

lay::Macro *MacroCollection::create (const char *prefix, Macro::Format format)
{
  std::string name;
  int n = 0;
  do {
    name = (prefix ? prefix : "new_macro");
    if (n > 0) {
      name += "_" + tl::to_string (n);
    }
    if (! macro_by_name (name, format)) {
      break;
    }
    ++n;
  } while (true);

  begin_changes ();

  lay::Macro *m = m_macros.insert (std::make_pair (name, new lay::Macro ()))->second; 
  m->set_name (name);
  m->set_parent (this);

  on_changed ();

  return m;
}

bool MacroCollection::add (lay::Macro *m)
{
  QDir d (tl::to_qstring (path ()));
  QDir dd = QFileInfo (tl::to_qstring (m->path ())).dir ();

  if (d == dd) {

    begin_changes ();
    m_macros.insert (std::make_pair (m->name (), m))->second;
    m->set_parent (this);
    on_changed ();
    return true;

  } else {

    for (child_iterator c = begin_children (); c != end_children (); ++c) {
      if (c->second->add (m)) {
        return true;
      }
    }

    //  try to detect new child folders. If that is the case, create that folder and add
    //  the macro there.
    QDir dm (tl::to_qstring (m->dir ()));
    while (true) {

      std::string folder_name = tl::to_string (dm.dirName ());
      if (! dm.cdUp ()) {
        break;
      }

      if (dm == d) {
        begin_changes ();
        lay::MacroCollection *mc = m_folders.insert (std::make_pair (folder_name, new MacroCollection ())).first->second;
        mc->set_virtual (false);
        mc->set_parent (this);
        on_changed ();
        return mc->add (m);
      }

    }

  }

  return false;
}

bool MacroCollection::del ()
{
  if (tl::verbosity () >= 20) {
    tl::info << "Deleting macro folder " << path ();
  }
  return QDir ().rmdir (tl::to_qstring (path ()));
}

void MacroCollection::rename_macro (Macro *macro, const std::string &new_name)
{
  iterator m = m_macros.find (macro->name ());
  while (m != m_macros.end () && m->first == macro->name ()) {
    if (m->second == macro) {
      m_macros.erase (m);
      m_macros.insert (std::make_pair (new_name, macro));
      return;
    }
    ++m;
  }
}

lay::Macro *MacroCollection::find_macro (const std::string &path)
{
  for (iterator m = begin (); m != end (); ++m) {
    if (m->second->path () == path) {
      return m->second;
    }
  }

  for (child_iterator mc = begin_children (); mc != end_children (); ++mc) {
    lay::Macro *macro = mc->second->find_macro (path);
    if (macro) {
      return macro;
    }
  }

  return 0;
}

MacroCollection &MacroCollection::root ()
{
  return ms_root;
}

static bool has_autorun_for (const lay::MacroCollection &collection, bool early)
{
  for (lay::MacroCollection::const_child_iterator c = collection.begin_children (); c != collection.end_children (); ++c) {
    if (has_autorun_for (*c->second, early)) {
      return true;
    }
  }

  for (lay::MacroCollection::const_iterator c = collection.begin (); c != collection.end (); ++c) {
    if ((early && c->second->is_autorun_early ()) || (!early && c->second->is_autorun () && !c->second->is_autorun_early ())) {
      return true;
    }
  }

  return false;
}

bool MacroCollection::has_autorun () const
{
  return has_autorun_for (*this, false);
}

static void autorun_for (lay::MacroCollection &collection, bool early)
{
  for (lay::MacroCollection::child_iterator c = collection.begin_children (); c != collection.end_children (); ++c) {
    autorun_for (*c->second, early);
  }

  for (lay::MacroCollection::iterator c = collection.begin (); c != collection.end (); ++c) {
    if ((early && c->second->is_autorun_early ()) || (!early && c->second->is_autorun () && !c->second->is_autorun_early ())) {
      BEGIN_PROTECTED
        c->second->install_doc ();
        c->second->run ();
      END_PROTECTED
    }
  }
}

void MacroCollection::autorun ()
{
  autorun_for (*this, false);
}

void MacroCollection::autorun_early ()
{
  autorun_for (*this, true);
}

void MacroCollection::dump (int l)
{
  for (int i = 0; i < l; ++i) { printf ("  "); }
  printf ("----\n");
  for (int i = 0; i < l; ++i) { printf ("  "); }
  printf ("Collection: %s\n", name ().c_str ());
  for (int i = 0; i < l; ++i) { printf ("  "); }
  printf ("Collection-path: %s\n", path ().c_str ());
  for (int i = 0; i < l; ++i) { printf ("  "); }
  printf ("Collection-description: %s\n", description ().c_str ());
  for (int i = 0; i < l; ++i) { printf ("  "); }
  printf("Collection-readonly: %d\n", is_readonly ());
  printf ("\n");

  for (iterator m = begin (); m != end (); ++m) {
    for (int i = 0; i < l; ++i) { printf ("  "); }
    printf("Name: %s%s\n", m->second->name ().c_str (), m->second->is_modified() ? "*" : "");
    for (int i = 0; i < l; ++i) { printf ("  "); }
    printf("  Path: %s\n", m->second->path ().c_str ());
    for (int i = 0; i < l; ++i) { printf ("  "); }
    printf("  Readonly: %d\n", m->second->is_readonly ());
    for (int i = 0; i < l; ++i) { printf ("  "); }
    printf("  Autorun: %d\n", m->second->is_autorun ());
    for (int i = 0; i < l; ++i) { printf ("  "); }
    printf("  Autorun-early: %d\n", m->second->is_autorun_early ());
    for (int i = 0; i < l; ++i) { printf ("  "); }
    printf("  Description: %s\n", m->second->description ().c_str ());
  }

  for (child_iterator m = begin_children (); m != end_children (); ++m) {
    m->second->dump (l + 1);
  }
}

// ----------------------------------------------------------------------

MacroSignalAdaptor::MacroSignalAdaptor (QObject *parent, Macro *macro)
  : QObject (parent), mp_macro (macro)
{
  //  .. nothing yet ..
}

void MacroSignalAdaptor::run ()
{
BEGIN_PROTECTED
  if (mp_macro) {
    mp_macro->run ();
  }
END_PROTECTED
}

}

